;;;--------------------------------------------------------------------------
;;;  INFORMATION:
;;;--------------------------------------------------------------------------
;;;  
;;;  This model simulates the acquisition of object pronouns and is described
;;;  in Van Rij, J., Van Rijn, H., & Hendriks, P. (submitted). Cognitive 
;;;  Architectures and Language Acquisition: A Case Study in Pronoun
;;;  Comprehension.
;;; 
;;;  author:    Jacolien van Rij
;;;  date:      April 2008
;;;  address:	Center for Language and Cognition Groningen
;;;  			University of Groningen
;;;				P.O.Box 716, 9700AS Groningen
;;;				J.C.van.Rij@rug.nl
;;;
;;;  This model forms the basis for the Speech Rate model (Van Rij, Van Rij, & 
;;;  Hendriks, submitted), which simulates the performance of a group of child 
;;;  participants on sentences with normal and slowed-down speech rate.
;;;  
;;;--------------------------------------------------------------------------
;;;  TO DO:
;;;--------------------------------------------------------------------------
;;;
;;;  The model is implemented in ACT-R 6, but the old utility-function (P-CG)
;;;  must be used for this model: 
;;;     * replace utility-and-reward-1.lisp in modules\ 
;;;       with the file utility-and-reward.lisp in extras\old-utility\
;;;
;;;  Use (do-experiment n) to run n simulations.
;;;  Use (run-simulation n uni-n) to run the model for n sessions of which 
;;;  uni-n sessions use only unidirectional optimization (bidrectional optimi-
;;;  zation is disabled in the first uni-n sessions).
;;; 
;;;--------------------------------------------------------------------------


;;;--------------------------------------------------------------------------
;; Declaration global variables
;;;--------------------------------------------------------------------------

(defvar *inputs* '((disjoint pronoun) (coreferential reflexive)))

(defvar *t-limit* .7)
(defvar *t-max-s* .1)
(defvar *%Reflexives* 10)

(defvar *uni-feedback* t)
(defvar *verbose* t)

(defvar *case* 0)
(defvar *item* 0)

(defvar *output-file*)
(setq *output-file* "outputfile.csv")


;;;--------------------------------------------------------------------------
;;; Output functions
;;;--------------------------------------------------------------------------

;; Initialization outputfile

(with-open-file
    (poutfile *output-file*
                :if-exists :append
                :if-does-not-exist :create
                :direction :io)

    (write-string
        (format nil "~A~%"
            "case;item;original-input;optimal;output;state;input;correct;biot" )
    poutfile)
)


;; Print output

(defun output-current-goal ()
    (with-open-file
    	(poutfile *output-file*
    			    :if-exists :append
    			    :if-does-not-exist :create
    			    :direction :io)
        ;; Print casenr
        (write-string (format nil "~d;" *case*) poutfile)

        (write-string (format nil "~d;" *item*) poutfile)

        ;; Print goal slots original-output, optimal, output and state
    	(write-string (format nil "~A;"
                      (chunk-slot-value-fct (no-output (goal-focus)) 'original-input))
                      poutfile)

    	(write-string (format nil "~A;"
    		      (chunk-slot-value-fct (no-output (goal-focus)) 'optimal))
                      poutfile)

    	(write-string (format nil "~A;"
    			(chunk-slot-value-fct (no-output (goal-focus)) 'output))
                      poutfile)
        
    	(write-string (format nil "~A;"
    			(chunk-slot-value-fct (no-output (goal-focus)) 'state))
                      poutfile)


        ;; Print codes for input type, correct and bidirectional optimization
        ;; if pronoun, then print 1; if reflexives, then 0; else -1
        (if (equal (chunk-slot-value-fct (no-output (goal-focus)) 'original-input) 
                    (no-output (cadar *inputs*)))
                (write-string (format nil "~A;" "1") poutfile)
                (if (equal (chunk-slot-value-fct (no-output (goal-focus)) 'original-input) 
                            (no-output (cadadr *inputs*)))
                    (write-string (format nil "~A;" "0") poutfile)
                    (write-string (format nil "~A;" "-1") poutfile)))

        ;; Print correct
        ;; if <pronoun,disjoint> or <reflexive,coreferential>, then 1; else 0
        (if (or 
                (and (equal (chunk-slot-value-fct (no-output (goal-focus)) 'original-input) 
                            (no-output (cadar *inputs*))) ;; pronoun
                     (equal (chunk-slot-value-fct (no-output (goal-focus)) 'output) 
                            (no-output (caar *inputs*)))) ;; disjoint
                (and (equal (chunk-slot-value-fct (no-output (goal-focus)) 'original-input) 
                            (no-output (cadadr *inputs*))) ;; reflexive
                     (equal (chunk-slot-value-fct (no-output (goal-focus)) 'output) 
                            (no-output (caadr *inputs*)))) ;; coreferential
            )
            (write-string (format nil "~A;" "1") poutfile)
            (write-string (format nil "~A;" "0") poutfile))

        ;; Print bi-ot
        ;; if done_bi_pos or done_bi_neg, then 1; else 0
        (if (or (equal (chunk-slot-value-fct (no-output (goal-focus)) 'state) 
                   (no-output (first '(done_bi_pos done_bi_neg done_uni))))
                (equal (chunk-slot-value-fct (no-output (goal-focus)) 'state) 
                   (no-output (second '(done_bi_pos done_bi_neg done_uni)))))
            (write-string (format nil "~A~%" "1") poutfile)
            (write-string (format nil "~A~%" "0") poutfile)))
)


;;;--------------------------------------------------------------------------
;;; Functions for running model
;;;--------------------------------------------------------------------------

;; For another distribution of pronouns and reflexives, replace in do-one () 
;; the *%Reflexives*, which reflects the percentage of the input is reflexive.


;; Run one trial
(defun do-one ()
  (let*
    ((item (if (<= (act-r-random 100) *%Reflexives*) (nth 1 *inputs*) (nth 0 *inputs*))))
        (add-new-goal-to-DM (first item) (second item))
        (format *verbose* "~tpresenting: ~S - ~S~%" (first item) (second item))
        (reset-declarative-finsts)
        (run (+ *t-limit* (act-r-noise *t-max-s*)))
    	(setf temp (mod-chunk-fct (no-output (goal-focus)) (list 'time nil)))
        (format *verbose* "     ~toutput:~t~S~%~%" (chunk-slot-value-fct (no-output (goal-focus)) 'output))
    	(run 2)
    	(output-current-goal)
  )
)

;; Run many trails (from trialnr t1 to trialnr times, that is times-t1+1 trials)
(defun do-many (times &optional (t1 1))
      (setf *item* (1+ *item*))
      (and (>= times t1)
       (not (format *verbose* "~4d)" t1))
       (do-one)
       (do-many times (1+ t1)))
)

;; Run a simulation of the development of one person,
;; containing n trials of which the first uni-n are unidirectional
(defun run-simulation (n uni-n)
    (reset)
    (setf *item* 0)

    (if *uni-feedback* (penable-fct '(finished_reflexive)) (pdisable-fct '(finished_reflexive)))
    (pdisable-fct '(function1 function2))
    
    (format *verbose* "~%***~%UNIDIRECTIONAL OPTIMIZATION~%***~%" )
    (do-many uni-n)
    
    (penable-fct '(function1 function2))
    
    (format *verbose* "~%***~%BIDIRECTIONAL OPTIMIZATION~%***~%" )
    (do-many n (+ uni-n 1))
)

;; Run n independent simulations of 1000 trials
(defun do-experiment (n)
    (dotimes (i n)
      (setf *case* i)
      (format *verbose* "~%--------------------~%New subject: case ~d~%--------------------~%" i)
      (run-simulation 1000 200)
      )
)

;; Generate a new goal
(defun add-new-goal-to-DM (func inp)
    (let* ((top-goal (gentemp "language-fact")))
        (progn
            (eval `(add-dm
                (,top-goal ISA language-fact input ,inp function ,func state start time t)))
            (eval `(goal-focus ,top-goal))
            top-goal)))



;;;--------------------------------------------------------------------------
;;; Definition of the model (parameters and declaratie memory items)
;;;--------------------------------------------------------------------------
        
(clear-all)

(define-model acquisition_biOT

;;---------------------------------------------------------------------------        
;; Description parameters (see reference manual)

;; :ans n => Activation noise s parameter specifies the s value used to 
;;           generate the instantaneous noise added to the activation equation 
;;           (permanent activation noise van chunks)
;; :bll n => Base-level learning is enabled and the value of the decay parameter, 
;;           d, is set to n 
;; :declarative-finst-span n => Controls how long a finst can mark a chunk as 
;;           having been retrieved. After a finst has been on a chunk for this 
;;           amount of time it is removed and the chunk is no longer marked 
;;           as recently retrieved. (in this model (reset-declarative-finsts)
;;           is used to reset the finst at the start of every session)
;; :egs t => Utilities have noise added to them, which is controlled by the 
;;           utility noise parameter s which is set with the parameter :egs
;; :epl t => Production compilation is enabled
;; :esc t => Enable subsymbolic computations, specifies whether modules should 
;;           work in a purely symbolic fashion or not. If it is set to t, 
;;           then modules should use whatever subsymbolic computations they 
;;           provide 
;; :mas n =>
;; :pct t    Production compilation trace is enabled
;; :pl t =>  parameters learning, if enabled p and q will learned for each 
;;           production (parameter of old utility-function, in new function
;;           replaced with :ul)
;; :rt n =>  Retrieval threshold, the minimum activation a chunk must have to 
;;           be able to be retrieved, in the retrieval failure time equation. 
;;           (Very low :rt = less retrieval failures and faster retrieval time 
;;           in case of failure)
;; :ul t =>  Utility learning flag, if it is set to t then the utility learning 
;;           equation will be used to learn the utilities as the model runs.
;; :ut n =>  Utility threshold, the threshold for the utility of a production
;;           below which it will not be considered during conflict resolution
;;           (parameter of old utility-function)
;;---------------------------------------------------------------------------     

    ;; Parameters settings
    (sgp :v nil :trace-detail low :esc t :epl t :pct nil 
        :pl t :ut -100 :egs 1 
        :rt -100 :bll 0.5 :mas 2 :ans 0.3 
        :declarative-finst-span 3 
        :act nil :save-buffer-trace t :ncnar nil)

    ;; Definition chunk types
    (chunk-type language-fact input function cand1 cand2 original-input optimal output state process time)
    (chunk-type candidate cand input)
    (chunk-type constraint input ranking violates)


    ;; Declarative memory items
    (add-dm 
        ;; States
        (start isa chunk)           (get_candidate isa chunk)
        (getcand2 isa chunk)        (getoptimal isa chunk)
        (unsolved isa chunk)        (nextcandidate isa chunk)
        (function isa chunk)       (forced_output isa chunk)
        (done_uni isa chunk)        (done_bi_pos isa chunk)
        (done_bi_neg isa chunk)     (time-limit-reached isa chunk)

        ;; Concepts
        (interpretation isa chunk)  (production isa chunk)
        (first isa chunk) (second isa chunk) (third isa chunk)
        (pronoun isa chunk)         (reflexive isa chunk)
        (coreferential isa chunk)   (disjoint isa chunk)

        ;; Candidates - production
        (pro-co-prod isa candidate cand pronoun input coreferential)
        (pro-dis-prod isa candidate cand pronoun input disjoint)
        (self-co-prod isa candidate cand reflexive input coreferential)
        (self-dis-prod isa candidate cand reflexive input disjoint)
        ;; Candidates - interpretation
        (pro-co-int isa candidate cand coreferential input pronoun)
        (pro-dis-int isa candidate cand disjoint input pronoun)
        (self-co-int isa candidate cand coreferential input reflexive)
        (self-dis-int isa candidate cand disjoint input reflexive)
        ;; Dummy candidate
        (no-candidate isa candidate cand nil input nil)

        ;; Constraint - Principle A 
        (principle_A_refl_int isa constraint input reflexive ranking first violates disjoint)
        (principle_A_pro_int isa constraint input pronoun ranking first violates nil)  
        (principle_A_dis_prod isa constraint input disjoint ranking first violates reflexive)
        (principle_A_co_prod isa constraint input coreferential ranking first violates nil)
        ;; Constraint - Refential Economy 
        (avoidPronouns_refl_int isa constraint input reflexive ranking second violates nil)
        (avoidPronouns_pro_int isa constraint input pronoun ranking second violates nil)
        (avoidpronouns_co_prod isa constraint input coreferential ranking second violates pronoun)
        (avoidPronouns_dis_prod isa constraint input disjoint ranking second violates pronoun) 
        ;; Dummy constraint
        (no-constraint isa constraint input nil ranking third violates nil)
    )

    ;; Set baselevel actiation, by defining creation time and number of references
    (sdp pronoun :creation-time -1000000 :reference-count 75000)
    (sdp reflexive :creation-time -1000000 :reference-count 25000)
    (sdp disjoint :creation-time -1000000 :reference-count 50000)
    (sdp coreferential :creation-time -1000000 :reference-count 50000)

    (sdp pro-dis-int :creation-time -50000 :reference-count 500)
    (sdp pro-co-int :creation-time -50000 :reference-count 500)
    (sdp self-dis-int :creation-time -50000 :reference-count 500)
    (sdp self-co-int :creation-time -50000 :reference-count 500)

    (sdp pro-dis-prod :creation-time -50000 :reference-count 500)
    (sdp pro-co-prod :creation-time -50000 :reference-count 500)
    (sdp self-dis-prod :creation-time -50000 :reference-count 500)
    (sdp self-co-prod :creation-time -50000 :reference-count 500)

    (sdp no-candidate :creation-time -50000 :reference-count 1000)

    (sdp principle_A_refl_int :creation-time -50000 :reference-count 750)
    (sdp principle_A_pro_int :creation-time -50000 :reference-count 1000)
    (sdp principle_A_dis_prod :creation-time -50000 :reference-count 850)
    (sdp principle_A_co_prod :creation-time -50000 :reference-count 850)

    (sdp avoidPronouns_refl_int :creation-time -50000 :reference-count 500)
    (sdp avoidPronouns_pro_int :creation-time -50000 :reference-count 750)
    (sdp avoidpronouns_co_prod :creation-time -50000 :reference-count 650)
    (sdp avoidPronouns_dis_prod :creation-time -50000 :reference-count 650)

    (sdp no-constraint :creation-time -50000 :reference-count 1000)

    ;;Sji: This is the strength of association from source j to chunk i. 
    (add-sji (pronoun principle_A_pro_int 15)(principle_A_pro_int pronoun 15))
    (add-sji (pronoun avoidPronouns_pro_int 10)(avoidPronouns_pro_int pronoun 10))
    (add-sji (reflexive principle_A_refl_int 7)(principle_A_refl_int reflexive 7))
    (add-sji (reflexive avoidPronouns_refl_int 5)(avoidPronouns_refl_int reflexive 5))

    (add-sji (disjoint principle_A_dis_prod 10)(principle_A_dis_prod disjoint 10))
    (add-sji (coreferential principle_A_co_prod 7)(principle_A_co_prod coreferential 7))
    (add-sji (disjoint avoidPronouns_dis_prod 10)(avoidPronouns_dis_prod disjoint 10))
    (add-sji (coreferential avoidPronouns_co_prod 5)(avoidPronouns_co_prod coreferential 5))

    (add-sji (pro-dis-int pronoun 3) (pronoun pro-dis-int 3))
    (add-sji (pro-dis-int disjoint 3) (disjoint pro-dis-int 3))
    (add-sji (pro-co-int pronoun 3) (pronoun pro-co-int 3))
    (add-sji (pro-co-int coreferential 3) (coreferential pro-co-int 3))
    (add-sji (self-dis-int reflexive 3) (reflexive self-dis-int 3))
    (add-sji (self-dis-int disjoint 3) (disjoint self-dis-int 3))
    (add-sji (self-co-int reflexive 3) (reflexive self-co-int 3))
    (add-sji (self-co-int coreferential 3) (coreferential self-co-int 3))

    (add-sji (pro-dis-int pronoun 3) (pronoun pro-dis-prod 3))
    (add-sji (pro-dis-prod disjoint 3) (disjoint pro-dis-prod 3))
    (add-sji (pro-co-prod pronoun 3) (pronoun pro-co-prod 3))
    (add-sji (pro-co-prod coreferential 3) (coreferential pro-co-prod 3))
    (add-sji (self-dis-prod reflexive 3) (reflexive self-dis-prod 3))
    (add-sji (self-dis-prod disjoint 3) (disjoint self-dis-prod 3))
    (add-sji (self-co-prod reflexive 3) (reflexive self-co-prod 3))
    (add-sji (self-co-prod coreferential 3) (coreferential self-co-prod 3))


;;;--------------------------------------------------------------------------
;;; Definition of the model (production rules)
;;;--------------------------------------------------------------------------

;; If an utterance is being met, start interpreting by retrieving a relevant candidate
(P start_int
	=goal>
		isa			    language-fact
        input		    =input
		function	    =function
        state			start
		process		    nil
		time			t
==>
	=goal>
        original-input	=input
        state			get_candidate
		process			interpretation
	+retrieval>
		isa			    candidate
		input		    =input
		:recently-retrieved	nil
)

;; If an utterance must be produced, start by retrieving a relevant candidate
(P start_prod
	=goal>
		isa			    language-fact
        input		    =input
		function	    =function
		original-input	=or_input
		state			start
		process			production
		time			t		
==>
	=goal>
		state			get_candidate
	+retrieval>
		isa			    candidate
		input		    =input
)

;; If one candidate is being retrieved, retrieve a second candidate
(P cand1
	=goal>
		isa			    language-fact
		input		    =input
        state			get_candidate
		process			=process
		time			t	
	=retrieval>
		isa 			candidate
		cand			=cand
==>
	=goal>
		cand1			=cand
		state			getcand2
	+retrieval>
		isa			    candidate
		input		    =input
		- cand			=cand
		:recently-retrieved	nil
)

;;If two candidates are being retrieved, retrieve the first constraint
(P cand2
	=goal>
		isa			    language-fact
		input		    =input
		cand1			=val
		state			getcand2
		time			t	
	=retrieval>
		isa 			candidate
		cand			=cand
==>
	=goal>
        
		cand2			=cand
		state			getoptimal
	+retrieval>
		isa			    constraint
        input		    =input
        ranking		    first
)


;; If the both retrieved candidates are the same, replace one
(P evaluation_same
	=goal>
		isa			    language-fact
		input		    =input
		cand1			=val1
		cand2			=val1 
		state			getoptimal
		time			t	
	=retrieval>
		isa 			constraint
		ranking			first
==>
	=goal>
		cand1			=val1
		cand2			nil 
		state			getcand2
	+retrieval>
		isa			    candidate
		input		    =input
		- cand			=val1
)

(spp  evaluation_same :failure t)


;; If the first constraint does not violate one of the candidates, 
;; retrieve the second constraint
(P evaluation_eq1
	=goal>
		isa			    language-fact
		input		    =input
		cand1		    =val1
		cand2		    =val2
		state		    getoptimal
		time		    t	
	=retrieval>
		isa 		    constraint
		ranking		    first
        violates        nil 
==>
	=goal>
		state           getoptimal
	+retrieval>
		isa             constraint
		input           =input        
        ranking         second
)


;; If the second constraint does not violate one of the candidates, 
;; retrieve the third constraint 
(P evaluation_eq2
	=goal>
		isa			    language-fact
		input		    =input
		cand1			=val1
		cand2			=val2
		state			getoptimal
		time			t	
	=retrieval>
		isa 			constraint
		ranking			second
        violates        nil
==>
	=goal>
		state			unsolved
	+retrieval>
		isa			    constraint
        ranking		    third
)		

;; If the constraint violates the first candidate, 
;; mark the second candidate as optimal and retrieve the next candidate
(P evaluation_diff
	=goal>
		isa			    language-fact
		input		    =input
		cand1			=violates
		cand2			=val2
		state			getoptimal
		process			=process        
		time			t	
	=retrieval>
		isa 			constraint
        violates        =violates
==>
	=goal>
		optimal			=val2
		state			nextcandidate
	+retrieval>
		isa			    candidate
        input           nil
		- cand			=val2
		- cand			=violates
)

;; If the constraint violates the second candidate, 
;; mark the first candidate as optimal and retrieve the next candidate
(P evaluation_diff2
	=goal>
		isa			    language-fact
		input		    =input
		cand1			=val1
		cand2			=violates
		state			getoptimal
        process			=process
		time			t	
	=retrieval>
		isa 			constraint
        violates        =violates
==>
	=goal>
		optimal			=val1
		state			nextcandidate
	+retrieval>
		isa			    candidate
        input           nil
		- cand			=val1
		- cand			=violates
)

;; If no candidate can be retrieved at all an output cannot be given
(P no_candidate1
	=goal>
		isa			    language-fact
		input		    =input
		state			get_candidate
	?retrieval>
		state			error
==>
	=goal>
		optimal 		nil
		state			forced_output
)

;; If a second candidate cannot be retrieved, 
;; mark the first candidate as being optimal. Note that this is a forced choice! 
(P no_candidate2
	=goal>
		isa			    language-fact
		cand1			=cand
		state			getcand2
	?retrieval>
		state			error
==>
	=goal>
		optimal 		=cand
		state			done_uni
)

;; If a constraint cannot be retrieved, mark an arbitrary candidate as being optimal. 
;; This is also a forced choice situation! 
(P no_constr12
	=goal>
		isa			    language-fact
		cand1			=cand1
		cand2			=cand2
		state			getoptimal
	?retrieval>
		state			error
==>
	=goal>
		optimal			=cand1
		state			done_uni
)

;; If no next candidate can be retrieved, the unidirectional optimal interpretation is found 
(P no_next_candidate
	=goal>
		isa			    language-fact
		optimal			=optimal
		state			nextcandidate
        process			interpretation
		time			t
	=retrieval>
		isa             candidate
        input           nil       
==>
	=goal>
		state			done_uni
)

;; If there are no constraints left to be retrieved, 
;; mark an arbitrary candidate as being optimal and retrieve a next candidate
;; candidate 2 is marked as optimal candidate
(P no_next_constraint
	=goal>
		isa			    language-fact
		input		    =input
		cand1			=val1
		cand2			=val2
		state			unsolved
        process			=process
		time			t	
	=retrieval>
		isa             constraint
        input           nil
==>
	=goal>
		optimal			=val2
		state			nextcandidate
	+retrieval>
		isa			    candidate
        input           nil
		- cand			=val1
		- cand			=val2
)

;; If there are no constraints left to be retrieved, 
;; mark an arbitrary candidate as being optimal and retrieve a next candidate
;; candidate 1 is marked as optimal candidate
(P no_next_constraint2
	=goal>
		isa			    language-fact
		input		    =input
		cand1			=val1
		cand2			=val2
		state			unsolved
        process			=process
		time			t	
	=retrieval>
		isa             constraint
        input           nil
==>
	=goal>
		optimal			=val1
		state			nextcandidate
	+retrieval>
		isa			    candidate
        input           nil
		- cand			=val1
		- cand			=val2
)

;; If no next candidate can be retrieved, the unidirectional optimal interpretation is found. 
;; Find the bidirectional interpretation
(P bidirectional
	=goal>
		isa			    language-fact
        input           =input 
		function	    =function
		optimal			=optimal
		state			nextcandidate
        process			interpretation
		time			t	
	=retrieval>
		isa             candidate
        input           nil   
==>
	=goal>
		function 	    =optimal
		state			function
        output          =optimal
        !output! (Finished unidirectional process)
)

;; If the unidirectional optimal interpretation is found, 
;; Find the bidirectional optimal interpretation 
(P bidirectional2
	=goal>
		isa			    language-fact
        input           =input
        function		=function
		optimal			=optimal
		state			done_uni
        process			interpretation
		time			t	

==>
	=goal>
		function        =optimal
		state           function
        output          =optimal
        !output! (Finished unidirectional process)
)

;; feedback reflexieven
(P finished_reflexive
	=goal>
		isa			    language-fact
        input           reflexive
		function		=function
		optimal			=optimal
		state			nextcandidate
        process			interpretation
		time			t	
	=retrieval>
		isa             candidate
        input           nil   
==>
	=goal>
		function 		=optimal
        output          =optimal
		state			function
        !output! (Finished unidirectional process)
)

;; If you were interpreting, start producing
(P function1
	=goal>
		isa			    language-fact
        input           =input
		function		=function
		optimal			=optimal
		state			function
        process         interpretation
		time			t	
==>
	=goal>
		input		    =optimal
        function        =input
        state			start        
		process			production	
)

;; If you were producing, start interpreting
(P function2
	=goal>
		isa			    language-fact
        input           =input
		function	    =function
		optimal			=optimal
		state			function
        process         production
		time			t	
==>
	=goal>
        input		    =optimal
        function		=input
		state			start
		process			interpretation
)

;; If no next candidate can be retrieved and the output is equal 
;; to the original input, than the interpretation is correct
(P output_bi_pos
	=goal>
		isa			    language-fact		
        original-input	=val
		optimal			=val
		state			nextcandidate
        process 		production
		time			t	
	=retrieval>
		isa             candidate
        input           nil 
==>
	=goal>
		state			done_bi_pos
                
)
                                                                        
;; If no next candidate can be retrieved and the output is not equal 
;; to the original input, than the interpretation is not correct
(P output_bi_neg
	=goal>
		isa			    language-fact
	  - original-input	=val
        optimal			=val
		state			nextcandidate
        process 		production
		time			t	
	=retrieval>
		isa             candidate
        input           nil 
==>
	=goal>
		state			done_bi_neg
)                                                                      

;; If the timelimit is reached, stop processing 
(P time-limit-reached
	=goal>
		isa 			language-fact
		time			nil
==>
	=goal>
		state			time-limit-reached
		time			time-limit-reached

)

;; FEEDBACK

;; interpretation for reflexive found
(spp finished_reflexive :success t)
;; Success in bidirectional optimization
(spp output_bi_pos :success t)  
;; Failure in bidirectional optimization
(spp output_bi_neg :failure t)    

)
